{
module E = Errormsg
module I = Input

open Lfparse

let lexerror msg lexbuf = E.parse_error msg

let scan_escape str =
  match str with
    "n" -> "\n"
  | "r" -> "\r"
  | "t" -> "\t"
  | "b" -> "\b"
  | "f" -> "\012"  (* ASCII code 12 *)
  | "v" -> "\011"  (* ASCII code 11 *)
  | "a" -> "\007"  (* ASCII code 7 *)
  | _ -> str


}

let digit    = ['0'-'9']
let letdig   = ['0'-'9' 'a'-'z' 'A'-'Z' '_' '-' '$' '#' '@' '!' '+'
                '<' '>' '?' '\'' '~' '.' ]

let letter    = ['a'-'z' 'A' - 'Z' '$' '_']
let ws       = [' ' '\009' '\012']
let pathname = ['a'-'z' 'A'-'Z' '0'-'9' '.' '/' '\\' '-']

let escape = '\\' _

rule skiptolf = parse
|  "ANN_STARTLF"        { STARTLF }
|  '\n'                 { begin
			    E.newline ();
			    skiptolf lexbuf 
			  end }
|   eof                 { EOF }
|   _                   { skiptolf lexbuf }

and token = parse
    '\n'		{ begin
			    E.newline ();
			    token lexbuf 
			  end }
  | "ANN_ENDLF"         { skiptolf lexbuf }

  | "%!Decl"		{ BEGINDECL }
  | "%!EndDecl"		{ ENDDECL }
  | "%!COMM"[^'!''\n']*'\n'
      			{ begin
			    E.newline ();
			    let comm = Lexing.lexeme lexbuf in
			      Comment (String.sub comm 0 
                                         ((String.length comm) - 1))
			  end }
  | '%'         	{ endline lexbuf }

  | "IMPORT"[^'\n']*'\n' { begin
                            E.newline ();
                            token lexbuf
                          end}
  | ws			{ token lexbuf }
  | '('			{ LPAREN }
  | ')'			{ RPAREN }
  | '['			{ LBRACKET }
  | ']'			{ RBRACKET }
  | '{'			{ LBRACE }
  | '}'			{ RBRACE }
  | "->"		{ ARROW }
  | '.'			{ DOT }
  | ':'			{ COLON }
  | ','                 { COMMA }
  | '+'			{ PLUS }
  | '*'			{ TIMES }
  | "<<"		{ Id "<<" }
  | ">>"		{ Id ">>" }
  | "<"			{ LT }
  | ">"			{ GT }
  | "<="		{ LEQ }
  | ">="		{ GEQ }
  | '='			{ EQ }
  | "<>"		{ NEQ }
  | "mod"		{ MOD }
  | "div"		{ DIV }
  | "floor"		{ FLOOR }
  | "ceil"		{ CEIL }
  | "true"		{ TRUE }
  | "false"		{ FALSE }
  | "block"             { BLOCK }
  | "/\\"		{ AND }
  | "\\/"		{ OR }
  | "=>"		{ IMPL }
  | "all"		{ ALL }
  | "exp"		{ EXP }
  | "pred"		{ PRED }
  | "pf"		{ PF }
  | "choose"            { CHOOSE }
  | ['-' '~' ]?(digit)+	{ let str = Lexing.lexeme lexbuf in
			  let len = String.length str in
			  let zero = Char.code '0' in
                          assert (len > 0);
			  let rec accum a d =
			    let acc c = a + (d * ((Char.code c) - zero)) in
			    function
			      0 -> let c = str.[0] in
				   if c='~' || c='-' then - a else (acc c)
			    | i -> accum (acc str.[i]) (d * 10) (i - 1)
			  in
			  Num (accum 0 1 (len-1)) }
  | '-'			{ MINUS }
  | '"'			{ (* '"' *)
                            let s = str lexbuf in 
                            try STRING s
                            with e -> 
                              raise (Failure 
                                        ("str: " ^ Printexc.to_string e)) 
                        } 
  | (letter)letdig*	{ Id (Lexing.lexeme lexbuf) }
  | eof			{ EOF }

  | _			{ 
                          begin
                            lexerror ("Illegal Character '" ^ 
                                      (Lexing.lexeme lexbuf)) lexbuf;
			    token lexbuf
			  end }

and endline = parse 
        '\n' 		{ E.newline (); token lexbuf}
|       eof             { EOF }
|	_		{ endline lexbuf}

and str = parse	
        '"'             { ""} (* '"' *)

|	escape		{let this = Lexing.lexeme lexbuf in 
                         let rest = str lexbuf in
                         let cur = scan_escape (String.sub this 1 1) in
                         cur ^ rest }

|	_		{let cur = Lexing.lexeme lexbuf in 
                         cur ^  (str lexbuf)} 

{

module S = Satproc

let initParser fname lexbuf = 
  match skiptolf lexbuf with 
    Lfparse.STARTLF -> ()
  | _ -> E.s (E.error "I cannot find ANN_STARTLF in the file %s\n" 
                fname)  

let parseNextFewInputs lexbuf =
  Parsing.clear_parser ();
  [ Lfparse.main token lexbuf ]

let parser: S.parserSpec = 
    { S.language = "lf";
      S.initParser  = initParser;
      S.nextFewInputs = parseNextFewInputs;
    }

} 
