(** This file contains the definition of a language of expressions and 
 * predicates. It is supposed to be used both for the Kettle theorem prover 
 * and for the Open Verifier *)


open Pretty
module E = Errormsg
module H = Hashtbl
module U = Util

type location = E.location

(** {b Expressions} *)

(** The main type for symbolic expressions. Each expression can have one of a 
 * few types: 
 - Int: any integer value
 - Sign32: a value that can be represented in 2's complement on 32 bits 
   (ranging from -2^31..2^31-1)
 - Unsign32: a value ranging from 0..2^32-1
 - Other: most notably the values of the "upd" memory constructor
*)
type var = string

type exp =                             
  | Int of int32                        (** Integer constants (of type 
                                         * Sign32). If you want to convert to 
                                         * an int, use {!Logic.to_int}, not 
                                         * In32.to_int because the latter 
                                         * silently drops a bit. *)

  | UnOp of unop * exp                  (** A unary operation. Type of 
                                         * operand and result depends on the 
                                         * operation. *)
  | BinOp of binop * exp * exp          (** A binary operation. Type of 
                                          * operand and result depends on the 
                                          * operation. **)

  | Fun of string * exp list		(** A constructed expression, with 
                                         * the constructor name and a number 
                                         * of arguments. *)

  | Var of int (*de Bruijn*)            (** A logical variable. This 
                                         * expression can only occur within 
                                         * the scope of a forall or exist 
                                         * predicate. The integer is a 
                                         * deBruijn index. *)

  | NVar of var                          (** A named variable *)
  | UVar of int                          (** A unification variable *)

and unop = 
    Mod32                               (** Computes modulo 2^32. Of type Int 
                                         * -> Unsign32. *)
  | BNot                                (** Bitwise complement *)
  | RepS32                              (** Tests whether the argument (an 
                                         * Int) is in the range of Sign32. 
                                         * Has a value of zero or one *)
  | RepU32                              (** Tests whether the argument (an 
                                         * Int) is in the range of Unsign32. 
                                         * Has a value of zero or one *)

and binop = 
    Plus                                (** Addition (the real one). Two 
                                         * operands of type Int or Sign32 and 
                                         * result of type Int.  *)
  | Minus                               (** Subtraction (the real one). Two 
                                         * operands of any integer type and 
                                         * result of type Int. *)
  | Times                               (** Multiplication (the real one ). 
                                         * Two operands of type Int or Sign32 
                                         * and result of type Int. *)
 
  (** The following operators work on 32-bit integers *)
  | Add                                 (** Addition modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Sub                                 (** Subtraction modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Mult                                (** Multiplication modulo 2^32. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)
  | Div                                 (** Integer Division. Two 
                                         * operands of type Sign32 and result 
                                         * of type Sign32. *)

  | BAnd                                (** Bitwise AND *)
  | BOr                                 (** Bitwise OR *)
  | BXor                                (** Bitwise XOR *)
      
  | Sll                                 (** Shift left logical *)
  | Srl                                 (** Shift right logical *)
  | Sra                                 (** Shift right arithmetic *)

  (** The following operators compare two 32-bit values and result in either 
    * one or zero *)
  | SetEq                               (** equal, printed "=" when used with 
                                         * PNez. *)
  | SetNe                               (** not equal, printed "<>" when used 
                                         * with PNez. *)
  | SetGt                               (** signed greater, printed ">" when 
                                         * used with PNez. *)
  | SetGe                               (** signed greater or equal, printed 
                                         * ">=" when used with PNez. *)
  | SetLt                               (** signed less, printed "<" when 
                                         * used with PNez. *)
  | SetLe                               (** signed less or equal, printed 
                                         * "<=" when used with PNez. *)
  | SetAb                               (** unsigned greater (above), printed 
                                         * ">u" when used with PNez. *)
  | SetAe                               (** unsigned greate equal, , printed 
                                         * ">=u" when used with PNez.  *)
  | SetBl                               (** unsigned less (below), printed 
                                         * "<u" when used with PNez. *)
  | SetBe                               (** unsigned less or equal , printed 
                                         * "<=u" when used with PNez. *)
  | SetPe                               (** parity even *)
  | SetPo                               (** parity odd *)

  | SignExt                             (** signext(e1,e2) we sign-extend the 
                                         * least-significant e2 bytes of e1, 
                                         * to a full 32-bit word  *)

(** A symbolic predicate *)
type pred =
    Bool of bool                    (** A boolean value of True or False *)  
  | PNez of exp                     (** A shorthand for e <> 0 *)
  | PFun of string * exp list       (** Constructed predicate *)

  | And of pred list                (* A conjunction of predicates. *)
  | Or of pred list                 (* A disjunction of predicates. *)

  | Imply of pred * pred            (** An implication *)
  | All of string * pred            (** A universal quantification of 
				      * the variable and predicate. Give the 
                                      * name of the quantified variable. *)
  | Exist of string * pred         (** An existential quanitification of the 
                                    * variable and predicate. Give the name 
                                    * of the quantified variable. *)


let isAtomPred (p: pred) : bool = 
  match p with 
    Bool _ | PNez _ | PFun _ -> true
  | _ -> false



let  count = ref 0 
let  newVar name = 
  incr count;
  name ^ "_" ^ (string_of_int !count)
let  resetVars () = count := 0
let  varName v = 
  try
    String.sub v 0 (1 + String.rindex v '_')
  with Not_found -> E.s (E.bug "%s is not a variable name" v)

(* ------ Logic pretty printer ------ *)


(** A printer interface for Logic trees. Create instantiations of 
 * this type by specializing the class {!Logic.defaultLogicPrinterClass}. Each 
 * method returns None if it wants the default printing (actually, that of 
 * the superclass) *)
class type logicPrinter = object

  method pExp: exp -> Pretty.doc option
    (** Invoked for each expression. If it return None, it defaults to the 
     * default printer. *)

  method pPred: pred -> Pretty.doc option
    (** Invoked for each predicate *)

end

let binop_to_string (b : binop) = 
  match b with 
    Plus ->  "+"
  | Minus ->  "-"
  | Times ->  "*"
  | Add ->  "add"
  | Sub ->  "sub"
  | Mult ->  "mult"
  | Div ->  "div"
  | BAnd ->  "band"
  | BOr ->  "bor"
  | BXor ->  "bxor"
  | SetPo ->  "setpo"
  | SetPe ->  "setpe"
  | SetBe ->  "setbe"
  | SetBl ->  "setbl"
  | SetAe ->  "setae"
  | SetAb ->  "setab"
  | SetLe ->  "setle"
  | SetLt ->  "setlt"
  | SetGe ->  "setge"
  | SetGt ->  "setgt"
  | SetNe ->  "setne"
  | SetEq ->  "seteq"
  | Sra    ->  "sra"
  | Srl    ->  "srl"
  | Sll    ->  "sll"
  | SignExt ->  "signext"
 
(** Return None if this predicate should not be printed specially *)
let pbinop_to_string (b: binop) : string option = 
  match b with 
  | SetBe -> Some "<=u"
  | SetBl -> Some "<u"
  | SetAe -> Some ">=u"
  | SetAb -> Some ">u"
  | SetLe -> Some "<="
  | SetLt -> Some "<"
  | SetGe -> Some ">="
  | SetGt -> Some ">"
  | SetNe -> Some "<>"
  | SetEq -> Some "="
  | _ -> None
   
let unop_to_string (u : unop) = 
  match u with 
  | Mod32 ->  "mod32"
  | BNot ->  "bnot"
  | RepS32 ->  "repS32"
  | RepU32 ->  "repU32"

(** Sometimes I am running into confusing bugs because Band and Fun("band")
 * are printed in the same way *)
let printFunctionsSpecially = false

let rec printExp (lp: logicPrinter) (e: exp) : doc = 
  (** See what the printer says *)
  match lp#pExp e with 
    Some d -> d
  | None -> begin
      let pe e = printExp lp e in
      let pelist el = (seq ((*chr ',' ++*) break) pe el) in
      match e with
        Int(i) -> text (Int32.to_string i)

      | UnOp(uop, e1) -> 
          text "(" ++ text (unop_to_string uop) ++ text " " ++ 
            align ++ pe e1 ++ chr ')' ++ unalign
      | BinOp(bop, e1, e2) -> 
          text "(" ++ text (binop_to_string bop) ++ text " " 
            ++ align ++ pe e1 ++ 
	    break ++ pe e2 ++ chr ')' ++ unalign
      | Fun(str, []) -> text str
      | Fun(str, explist) -> chr '(' ++ 
            (if printFunctionsSpecially then 
              text ("." ^ str) 
            else
              text str) ++ align ++ break  ++
	    pelist explist ++ chr ')' ++ unalign
      | Var(v) -> text "dB" ++ num v
      | NVar v -> text v
      | UVar v -> text "UVar(" ++ num v ++ text ")"
  end


let rec printPred (lp: logicPrinter) (p: pred) : doc = 
  (** See what the pretty printer says *)
  match lp#pPred p with 
    Some d -> d
  | None -> begin
      let pp p = printPred lp p in
      let pplist pl = (seq ((*chr ',' ++*) break) pp pl) in
      let pe e = printExp lp e in
      let pelist el = (seq ((*chr ',' ++*) break) pe el) in
      match p with
        Bool true -> text "True"
      | Bool false -> text "False"
      | And(predlist) -> (text "(AND " ++ align ++ pplist predlist ++ 
			    text ")" ++ unalign )
      | Or(predlist) -> (text "(OR " ++ align ++ pplist predlist ++ 
		           text ")" ++ unalign )
      | Imply(p1, p2) -> (text "(=> " ++ align ++ pp p1 ++ break
			    ++ pp p2 ++ text ")" ++ unalign )
      | All(var, pred) -> (text "(All [" ++ align ++ text var ++ text "]" ++ 
			     break
			     ++ pp pred ++ text ")" ++ unalign )
      | Exist(var, pred) -> 
          (text "(Ex" ++ align ++ text "ist [" ++ text var ++ text "]" ++ 
	     break
	     ++ pp pred ++ text ")" ++ unalign )
      | PNez e -> begin
          match e with 
            BinOp(bop, e1, e2) -> begin
              match pbinop_to_string bop with 
                Some s -> 
                  dprintf "(@[%s %a@?%a@])" s
                    insert (pe e1) insert (pe e2)
              | None -> dprintf "(PNez %a)" insert (pe e)
        end
      | _ -> dprintf "(PNez %a)" insert (pe e)
      end
            
      | PFun(s, args) -> (text "(" ++ align ++ text s ++ break
			    ++ (pelist args) ++ text ")" ++ unalign )
  end


class defaultLogicPrinterClass: logicPrinter = object (self)
  (** Default to the usual stuff *)
  method pExp (e: exp) : Pretty.doc option =  None

  method pPred (pred: pred) : Pretty.doc option = None
end

let defaultLogicPrinter: logicPrinter = new defaultLogicPrinterClass

let d_exp () e = printExp defaultLogicPrinter e

let d_pred () p = printPred defaultLogicPrinter p


class infixLogicPrinterClass: logicPrinter = object (self)
  (** Default to the usual stuff *)
  method pExp (e: exp) : Pretty.doc option = 
    let pe e = printExp (self :> logicPrinter) e in
    match e with 
      BinOp(bop, e1, e2) -> 
        Some (text "(" ++ align ++ pe e1 ++ text " " ++ 
                text (binop_to_string bop) ++ text " " 
                ++ pe e2 ++ chr ')' ++ unalign)
    | _ -> None

  method pPred (pred: pred) : Pretty.doc option = 
    match pred with 
    | PNez (BinOp(bop, e1, e2)) -> begin
        let pe e = printExp (self :> logicPrinter) e in
        match pbinop_to_string bop with 
          Some s -> 
            Some (dprintf "(@[%a %s@?%a@])" 
                    insert (pe e1) s insert (pe e2))
        | None -> None
    end
    | _ -> None
end

let infixLogicPrinter: logicPrinter = new infixLogicPrinterClass


(* Convert an expression list to Pretty.doc *)
let d_explist () el = 
  seq (chr ',' ++ break) (d_exp ()) el

(** Some temporary wrappers until we fold the rest of the printing into the 
 * class *)	
let prn_exp e = d_exp () e
let prn_pred p = d_pred () p
let prn_predlist pl = 
  seq (chr ',' ++ break) (d_pred ()) pl
let prn_explist el = 
  seq (chr ',' ++ break) (d_exp ()) el


let d_loc () l = E.d_loc () l

(********* Pretty printing ***********)


let integer (i: int) : exp = 
  Int (Int32.of_int i)

let zero = integer 0
let one  = integer 1

(* replace with call to Util.Int32Op version *)
exception IntegerTooLarge
let to_int (i: int32) : int = 
  try U.Int32Op.to_int i
  with U.Int32Op.IntegerTooLarge -> raise IntegerTooLarge

let of_int (i: int) : int32 = Int32.of_int i

let minus e1 e2 = BinOp(Minus, e1, e2)
let plus e1 e2 = BinOp(Plus, e1, e2)
let times e1 e2 = BinOp(Times, e1, e2)

let eq e1 e2 = PNez(BinOp(SetEq, e1, e2))
let ne e1 e2 = PNez(BinOp(SetNe, e1, e2))
let geq e1 e2 = PNez(BinOp(SetGe, e1, e2))
let leq e1 e2 = PNez(BinOp(SetLe, e1, e2))
let gt e1 e2 = PNez(BinOp(SetGt, e1, e2))
let lt e1 e2 = PNez(BinOp(SetLt, e1, e2))

let nilMap = "nilMap"
let consMap = "consMap"

let rec insertMap (m: exp) (idx: int) (e: exp) : exp = 
  let idx' = Int32.of_int idx in
  let rec loop (m: exp) =
    match m with
      Fun(n, _) when n = nilMap -> Fun(consMap, [ integer idx; 
                                                  e; 
                                                  m ])
    | Fun(c, [Int i'; e'; m']) when c = consMap -> 
        if i' = idx' then 
          Fun(consMap, [Int idx'; e; m'])
        else if i' > idx' then 
          Fun(consMap, [Int idx'; e; m])
        else
          Fun(consMap, [Int i'; e'; loop m' ])
  
    | _ -> E.s (E.bug "insertMap not in a map")
  in
  loop m
          
let makeMap (il: (int * exp) list) : exp = 
  List.fold_right
    (fun (i,e) acc -> 
      insertMap acc i e)
    il
    (Fun(nilMap, []))

let findMap (m: exp) (i: int) = 
  let i' = Int32.of_int i in
  let rec loop = function
    Fun(n, []) -> raise Not_found
  | Fun(c, [Int i''; e''; m'']) -> 
      if i'' = i' then 
        e''
      else
        loop m''
  | _ -> E.s (E.bug "findMap not in a map")
  in
  loop m
      
(******************
 ******************  VISITOR
 ******************)
(** {b The visitor} *)

(** Different visiting actions. 'a will be instantiated with [exp], [pred] *)
type 'a visitAction = 
    SkipChildren                        (** Do not visit the children. Return 
                                            the node as it is. *)
  | DoChildren                          (** Continue with the children of this 
                                            node. Rebuild the node on return 
                                            if any of the children changes 
                                            (use == test) *)
  | ChangeTo of 'a                      (** Replace the expression with the 
                                            given one *)
  | ChangeDoChildrenPost of 'a * ('a -> 'a) (** First consider that the entire 
                                           exp is replaced by the first 
                                           parameter. Then continue with 
                                           the children. On return rebuild 
                                           the node if any of the children 
                                           has changed and then apply the 
                                           function on the node *)



(** A visitor interface for traversing Logic expressions and predicates. 
 * Create instantiations of this type by specializing the class 
 * {!Logic.nopLogicVisitor}. *)
class type logicVisitor = object

  method vexp: exp -> exp visitAction          
    (** Invoked on each expression occurrence. The subtrees are the 
     * subexpressions *)

  method vpred: pred -> pred visitAction          
    (** Invoked on each predicate occurrence. The subtrees are the 
     * subexpressions *)

end

(* the default visitor does nothing at each node, but does *)
(* not stop; hence they return true *)
class nopLogicVisitor : logicVisitor = object
  method vexp (e:exp) = DoChildren   (* expression *) 
  method vpred (p:pred) = DoChildren
end

(*** Define the visiting engine ****)
(* visit all the nodes in a Cil expression *)
let doVisit (vis: logicVisitor)
            (startvisit: 'a -> 'a visitAction) 
            (children: logicVisitor -> 'a -> 'a) 
            (node: 'a) : 'a = 
  let action = startvisit node in
  match action with
    SkipChildren -> node
  | ChangeTo node' -> node'
  | _ -> (* DoChildren and ChangeDoChildrenPost *)
      let nodepre = match action with
        ChangeDoChildrenPost (node', _) -> node'
      | _ -> node
      in
      let nodepost = children vis nodepre in
      match action with
        ChangeDoChildrenPost (_, f) -> f nodepost
      | _ -> nodepost


(* mapNoCopy is like map but avoid copying the list if the function does not 
 * change the elements. *)
(* We have one of these in Util, but at some point I thought that the Ocaml 
 * inliner works better if we bring this here. this is a performance critical 
 * routine! *)
let rec mapNoCopy (f: 'a -> 'a) = function
    [] -> []
  | (i :: resti) as li -> 
      let i' = f i in
      let resti' = mapNoCopy f resti in
      if i' != i || resti' != resti then i' :: resti' else li 


(* A visitor for lists *)
let doVisitList  (vis: logicVisitor)
                 (startvisit: 'a -> 'a list visitAction)
                 (children: logicVisitor -> 'a -> 'a)
                 (node: 'a) : 'a list = 
  let action = startvisit node in
  match action with
    SkipChildren -> [node]
  | ChangeTo nodes' -> nodes'
  | _ -> 
      let nodespre = match action with
        ChangeDoChildrenPost (nodespre, _) -> nodespre
      | _ -> [node]
      in
      let nodespost = mapNoCopy (children vis) nodespre in
      match action with
        ChangeDoChildrenPost (_, f) -> f nodespost
      | _ -> nodespost


let rec visitLogicExp (vis: logicVisitor) (e: exp) : exp = 
  doVisit vis vis#vexp childrenExp e
and childrenExp (vis: logicVisitor) (e: exp) : exp = 
  let vExp e = visitLogicExp vis e in
  match e with
    Int _ | Var _ | NVar _ | UVar _ -> e
  | UnOp(uo, e) -> 
      let e' = vExp e in
      if e' != e then UnOp(uo, e') else e
  | BinOp(bo, e1, e2) -> 
      let e1' = vExp e1 in let e2' = vExp e2 in
      if e1' != e1 || e2' != e2 then BinOp(bo, e1', e2') else e
  | Fun(n, el) -> 
      let el' = mapNoCopy vExp el in
      if el' != el then Fun(n, el') else e

let rec visitLogicPred (vis: logicVisitor) (p: pred) : pred = 
  doVisit vis vis#vpred childrenPred p
and childrenPred (vis: logicVisitor) (p: pred) : pred = 
  let vExp e = visitLogicExp vis e in
  let vPred p = visitLogicPred vis p in
  match p with 
    Bool _ -> p
  | PFun(n, el) -> 
      let el' = mapNoCopy vExp el in
      if el' != el then PFun(n, el') else p
  | PNez e -> 
      let e' = vExp e in
      if e' != e then PNez e' else p
  | Exist (v, p1) -> 
      let p1' = vPred p1 in
      if p1' != p1 then Exist (v, p1') else p
  | All (v, p1) -> 
      let p1' = vPred p1 in
      if p1' != p1 then All (v, p1') else p
  | Imply(p1, p2) -> 
      let p1' = vPred p1 in let p2' = vPred p2 in
      if p1' != p1 || p2' != p2 then Imply(p1', p2') else p
  | And pl ->
      let pl' = mapNoCopy vPred pl in
      if pl' != pl then And pl' else p
  | Or pl ->
      let pl' = mapNoCopy vPred pl in
      if pl' != pl then Or pl' else p


(****
 ****     SUBSTITUTIONS
 ****)
(** Substitution visitors *)
let rec substPred ?(depth:int=0) 
                  (doOneExp: int -> exp -> exp option) (p: pred) : pred = 
  let rec sp (depth: int) (p0: pred) = 
    let sp' (p: pred) = sp depth p in
    let se' (e: exp) = substExp ~depth:depth doOneExp e in
    match p0 with 
      Bool _ -> p0
    | And pl -> 
        let pl' = mapNoCopy sp' pl in
        if pl' != pl then And pl' else p0

    | Or pl ->
         let pl' = mapNoCopy sp' pl in
        if pl' != pl then Or pl' else p0

   | Imply (p1, p2) -> 
        let p1' = sp' p1 in
        let p2' = sp' p2 in
        if p1' != p1 || p2' != p2 then Imply (p1', p2') else p0

    | All (v, p) -> 
        let p' = sp (depth + 1) p in
        if p != p' then All (v, p') else p0
    | Exist (v, p) -> 
        let p' = sp (depth + 1) p in
        if p != p' then Exist (v, p') else p0

    | PNez e -> 
        let e' = se' e in 
        if e' != e then PNez e' else p0

    | PFun (n, el) -> 
        let el' = mapNoCopy se' el in
        if el' != el then PFun(n, el') else p0
  in
  sp depth p

and substExp ?(depth:int=0) 
             (doOneExp: int -> exp -> exp option) (e0: exp) : exp = 
  (* First check if we have a substitution *)
  match doOneExp depth e0 with 
    Some e' -> e'
  | None -> begin
      let se = substExp ~depth:depth doOneExp in
      match e0 with 
        Int _ | Var _ | UVar _ | NVar _ -> e0
      | UnOp (uo, e) -> 
          let e' = se e in
          if e' != e then UnOp(uo, se e) else e0

      | BinOp (bo, e1, e2) -> 
          let e1' = se e1 in
          let e2' = se e2 in
          if e1' != e1 || e2' != e2 then BinOp(bo, e1', e2') else e0

      | Fun (n, el) -> 
          let el' = mapNoCopy se el in
          if el' != el then Fun(n, el') else e0
  end


(****
 **** EXPLICIT SUBSTITUTIONS
 ****
 ****)

type soption =  Unif | Done of sexp 

and subst = soption array

and sexp = exp * subst

type spred = pred * subst

let emptySubst: subst = Array.create 0 Unif

let newSubst n :subst = Array.create n Unif

let newSubstFromList (el: exp list) = 
  let inits = List.map (fun e -> Done (e, emptySubst)) el in
  Array.of_list inits


  (* Unroll the top-level substitution *)
let rec unrollExp ((e,s) as orig) : sexp = 
  match e with 
    UVar i -> begin
      assert (i >= 0 && i < Array.length s);
      match s.(i) with
        Unif -> orig
      | Done es' -> unrollExp es'
    end
  | _ -> orig

(* Unroll completely. Returns a fully substituted expression, if all UVars 
 * are resolved. Returns None otherwise *)
let unrollUVarsIntern (s: subst) (e: exp) : exp option = 
  let hasUVars = ref false in
  let rec loopSubst (s: subst) (e: exp) = 
    let rec loop (e: exp) =
      match e with 
        (UVar i) -> begin
          assert (i >= 0 && i < Array.length s);
          match s.(i) with
            Unif -> hasUVars := true; e
          | Done (e', s') -> loopSubst s' e'
        end
      | (Int _ | Var _ | NVar _) -> e
      | BinOp(bop,e1,e2) -> 
          let e1' = loop e1 in
          let e2' = loop e2 in
          if e1' == e1 && e2' == e2 then e 
          else BinOp(bop, e1', e2')
              
      | UnOp(uop,e1) -> 
          let e1' = loop e1 in
          if e1' == e1 then e else UnOp(uop, e1')
      | Fun(n,al) -> 
          let al' = mapNoCopy loop al in
          if al' == al then e else Fun(n, al')
    in
    loop e
  in
  let e' = loopSubst s e in
  if !hasUVars then 
    None
  else
    Some e'

let unrollUVarsPredIntern (s: subst) (p: pred) : pred option = 
  let hasUVars = ref false in
  let rec loop (p: pred) = 
    match p with 
      Bool _ -> p
    | And pl  -> 
        let pl' = mapNoCopy loop pl in 
        if pl' == pl then p else And pl'

    | Or pl  -> 
        let pl' = mapNoCopy loop pl in 
        if pl' == pl then p else Or pl'

    | Imply (p1, p2) -> 
        let p1' = loop p1 in
        let p2' = loop p2 in
        Imply (p1', p2')

    | All (s, p1) -> 
        let p1' = loop p1 in
        if p1' == p1 then p else All (s, p1')
    | Exist (s, p1) -> 
        let p1' = loop p1 in
        if p1' == p1 then p else Exist (s, p1')
    | PNez e -> begin
        match unrollUVarsIntern s e with
          None -> hasUVars := true; p
        | Some e' -> 
            if e' == e then p else PNez e'
    end

    | PFun (n, args) -> 
        let args' = 
          mapNoCopy 
            (fun e -> 
              match unrollUVarsIntern s e with 
                  Some e' -> e'
              | _ -> e) args in
        if args' == args then 
          p
        else
          PFun(n, args')
  in
  let p' = loop p in
  if !hasUVars then 
    None
  else
    Some p'
  
let unrollUVars (s: subst) (e: exp) : exp = 
  match unrollUVarsIntern s e with 
    None -> e
  | Some e' -> e'

  (* Unroll completely and raise HasUvar if it has unification variables *)
exception HasUVar

let unrollUVarsRaise (s: subst) (e: exp) : exp = 
  match unrollUVarsIntern s e with
    None -> raise HasUVar
  | Some e' -> e'

let unrollUVarsPredRaise (s: subst) (p: pred) : pred = 
  match unrollUVarsPredIntern s p with 
    None -> raise HasUVar
  | Some p' -> p'
      
let unrollUVarsError (s: subst) (e: exp) : exp = 
  match unrollUVarsIntern s e with
    None -> E.s (E.error "unrollUVarsError")
  | Some e' -> e'

let unrollUVarsPredRaise (s: subst) (p: pred) : pred = 
  match unrollUVarsPredIntern s p with 
    None -> raise HasUVar
  | Some p' -> p'
 

(** Make a special printer, that understands UVar *)
class uvarLogicPrinterClass (subst: subst) : logicPrinter = object (self)
  inherit defaultLogicPrinterClass as super

  val subst = subst
    
  method pExp = function
      UVar i -> begin
        assert (i >= 0 && i < Array.length subst);
        match subst.(i) with
          Unif -> Some (dprintf "UVar(%d)" i)
        | Done (e', s') -> 
            (* Make a clone with a new subst *)
            Some (printExp ({< subst = s' >} :> logicPrinter) e')
      end
    | e -> None
end

let d_spred () ((p: pred), (s: subst)) = 
  printPred (new uvarLogicPrinterClass s) p

let d_sexp () ((e: exp), (s: subst)) = 
  printExp (new uvarLogicPrinterClass s) e


(* Print a substitution. Watch out for circular substitutions. *)
let printedSubsts: subst list ref = ref []
let rec d_substIntern (s: subst) : doc = 
  (* See if we have found this already. Return its index *)
  let rec mem = function 
      [] -> None
    | s' :: rest -> 
        if s == s' then 
          Some (List.length rest) 
        else
          mem rest
  in
  match mem !printedSubsts with 
    None -> (* Not already printed *)
      if Array.length s = 0 then 
        text "empty"
      else begin
        let sidx = List.length !printedSubsts in
        printedSubsts := s :: !printedSubsts;
        align ++
          (docArray ~sep:line
             (fun i so -> 
               text "S" ++ num sidx ++ text "[" ++ num i ++ text "] -> " ++ align ++ 
                 (match so with 
                   Unif -> text "Unif" 
                 | Done (e, s) -> d_sexpNoUnrollIntern (e, s))
                 ++ unalign)
             ()
             s) 
          ++
          unalign
      end
  | Some idx -> text "S" ++ num idx ++ text " (already printed)"

and d_sexpNoUnrollIntern (e, s) : doc = 
  d_exp () e ++ line ++
    text "with " ++ align ++ d_substIntern s ++ unalign

let d_subst () (s: subst) = 
  printedSubsts := [];
  d_substIntern s

let d_sexpNoUnroll () (e, s) = 
  printedSubsts := [];
  d_sexpNoUnrollIntern (e, s)


(*************************************************************************
 ****
 ****          UNIFICATION 
 ****
 **************************************************************************)
let rec eunify (changeSub: subst -> int -> sexp -> bool) 
               ((e1, s1) : sexp) ((e2, s2) : sexp) : bool = 
  let rec doE e1 e2 = 
    match e1, e2 with
      Int i1, Int i2 -> i1 = i2
    | Var v1, Var v2 -> v1 = v2
    | Fun(n1, a1), Fun(n2, a2) when n1 = n2 -> 
        let rec unifArgs alist1 alist2 = 
          match alist1, alist2 with
            [], [] -> true
          | aa1 :: rest1, aa2 :: rest2 -> 
              doE aa1 aa2 && unifArgs rest1 rest2
          | _, _ -> false
        in
        unifArgs a1 a2

    | BinOp(bop, e1, e1'), BinOp(bop', e2, e2') -> 
        bop = bop' && doE e1 e2 && doE e1' e2'
    | UnOp(uop, e1), UnOp(uop', e2) -> 
        uop = uop' && doE e1 e2
    | UVar i1, _ -> begin 
        let es2 = e2, s2 in
        assert (i1 >= 0 && i1 < Array.length s1);
        match s1.(i1) with
          Unif -> changeSub s1 i1 es2
        | Done es1' -> eunify changeSub es1' es2
    end
    | _, UVar i2 -> begin 
        let es1 = e1, s1 in
        assert (i2 >= 0 && i2 < Array.length s2);
        match s2.(i2) with
          Unif -> changeSub s2 i2 es1
        | Done es2' -> eunify changeSub es1 es2'
    end
    | _ -> false
  in
  doE e1 e2

let rec punify changeSub (a1, s1) (a2, s2) : bool = 
  let eu e1 e2 = eunify changeSub (e1, s1) (e2, s2) in
  let pu p1 p2 = punify changeSub (p1, s1) (p2, s2) in
  match a1, a2 with
  | PFun(n1, a1), PFun(n2, a2) -> 
      let rec unifArgs alist1 alist2 = 
        match alist1, alist2 with 
          [], [] -> true
        | [], _ -> false
        | _, [] -> false
        | aa1 :: rest1, aa2 :: rest2 -> 
            eu aa1 aa2 && unifArgs rest1 rest2
      in
      n1 = n2 && unifArgs a1 a2
  | PNez e1, PNez e2 -> eu e1 e2
  | Bool b1, Bool b2 -> b1 = b2
  | Or ps1, Or ps2 ->
      let rec punifArgs alist1 alist2 = 
        match alist1, alist2 with 
          [], [] -> true
        | [], _ -> false
        | _, [] -> false
        | aa1 :: rest1, aa2 :: rest2 -> 
            pu aa1 aa2 && punifArgs rest1 rest2
      in
      punifArgs ps1 ps2
  | And ps1, And ps2 ->
      let rec punifArgs alist1 alist2 = 
        match alist1, alist2 with 
          [], [] -> true
        | [], _ -> false
        | _, [] -> false
        | aa1 :: rest1, aa2 :: rest2 -> 
            pu aa1 aa2 && punifArgs rest1 rest2
      in
      punifArgs ps1 ps2
  | Imply (p1, q1), Imply (p2, q2) ->
      pu p1 p2 && pu q1 q2
  | Exist (v1, b1), Exist (v2, b2) ->
      v1 = v2 && pu b1 b2
  | All (v1, b1), All (v2, b2) ->
      v1 = v2 && pu b1 b2
  | _ -> false



(*****)
let negateExp (e: exp) : exp = 
  let negateBinop = function
      SetEq -> SetNe
    | SetNe -> SetEq
    | SetGt -> SetLe
    | SetLe -> SetGt
    | SetGe -> SetLt
    | SetLt -> SetGe
    | SetAb -> SetBe
    | SetBe -> SetAb
    | SetAe -> SetBl
    | SetBl -> SetAe
    | SetPe -> SetPo
    | SetPo -> SetPe
    | bop -> raise Exit
	(* E.s (E.bug "negateBinOp not defined for %a" d_binop bop) *)
  in
  match e with 
    BinOp(bop,e1,e2) -> 
      begin try BinOp(negateBinop bop, e1, e2)
      with Exit -> BinOp(SetEq,e,zero)
      end
  | _ -> BinOp(SetEq,e,zero) 
      (* E.s (E.bug "negateExp not defined for %a" d_exp e) *)

exception CannotNegate of pred
let negatePredRaise (p: pred) : pred = 
  match p with 
    PNez e -> PNez (negateExp e)
  | Bool b -> Bool (not b)
  | _ -> raise (CannotNegate p)

let negatePred (p: pred) : pred = 
  try negatePredRaise p 
  with CannotNegate _ -> E.s (E.unimp "negatePred %a" d_pred p)



(* Simplify an expression to an integer *)
let simplifyArith (e: exp) : int option = 
  (* Map non-arithmetic expression to a factor *)
  let hash : (exp, int ref) H.t = H.create 13 in
  let constant = ref 0 in
  
  let updateHash (e: exp) (factor: int) : unit = 
    try
      let pf = H.find hash e in
      pf := !pf + factor
    with Not_found -> 
      H.add hash e (ref factor)
  in
  let rec loop (factor: int) = function
      Int c -> constant := !constant + (to_int c) * factor
    | BinOp(Add, e1, e2) -> loop factor e1; loop factor e2
    | BinOp(Sub, e1, e2) -> loop factor e1; loop (- factor) e2
    | BinOp(Mult, Int f, e) 
    | BinOp(Mult, e, Int f) -> loop ((to_int f) * factor) e
    | e -> updateHash e factor
  in
  loop 1 e;
  let good = ref true in
  H.iter
    (fun e' i -> 
      if !i <> 0 then begin
        ignore (E.log
                  "Expression %a has factor %d when checking %a == 0\n" 
                  d_exp e' !i d_exp e);
        good := false;
      end)
    hash;
  if !good then 
    Some !constant
  else
    None

    
(* rename a variable *)
class renameVisitor (oldName: string) (newName: string) : logicVisitor = object
  inherit nopLogicVisitor
  
  method vexp (e:exp) : exp visitAction =
  begin
    match e with
    | NVar(n) when n=oldName -> ChangeTo(NVar(newName))
    | _ -> DoChildren
  end

  method vpred (p:pred) : pred visitAction =
  begin
    (* don't dig inside hiding bindings *)
    match p with
    | All(n,_) when n=oldName -> SkipChildren
    | Exist(n,_) when n=oldName -> SkipChildren
    | _ -> DoChildren
  end
end

let renameVariable (src: pred) (oldName: string) (newName: string) : pred =
begin
  let vis:logicVisitor = (new renameVisitor oldName newName) in
  (visitLogicPred vis src)
end
