open Pretty 
module E = Errormsg
module I = Input
module L = Logic
module S = Satproc
module Eng = Engine
module P = Proof
module U = Util

 (* The top-level function of the prover *)

                                        (* These are the decision procedures. 
                                         * To avoid using some of them you 
                                         * can use the -exclude comand line 
                                         * argument *)


(* The module Features is produce by the Makefile based on ./configure *)  
let allSatProcs = Features.allSatProcs

let debug = false

let proofChannel: out_channel ref = ref stderr


(*** LFi proof generator ****)
type pstate = 
    Assume
  | Conjunct of int              (* How many unclosed andi there are *)

type lfpgen = 
    { mutable proofs: pstate list;
    } 

let lfstate =                           (* The global state of the LFi proof 
                                         * generator *)
  { proofs = [];
  } 
let indentProofs = ref false


(**************** LF proofs ******************)
let lfiProofGen : Eng.proofgen = 
  let startVC name = 
    lfstate.proofs <- [Conjunct 0];
    let name' = 
      if String.length name >= 4 && String.sub name 0 3 = "Sat" then
        String.sub name 3 (String.length name - 3)
      else
        name
    in
    ignore (fprintf !proofChannel
              "\nANN_PROOF(%s,@! %%LF_" name')
  in
  let printLF thunk = 
    let olf = !I.printLFSyntax in
    I.printLFSyntax := true;
    let opd = !printDepth in
    printDepth := 999999;
    let opi = !printIndent in
    printIndent := !indentProofs;
    thunk ();
    I.printLFSyntax := olf;
    printDepth := opd;
    printIndent := opi
  in

  let assume (name: U.symbol) (p: L.pred) = 
    printLF 
      (fun () -> 
        ignore (fprintf !proofChannel
                  "(andi (impi [%s : pf @[%a@]]@!" (U.symbolName name) 
                  L.d_pred p));
    match lfstate.proofs with
      Conjunct n :: rest -> 
        lfstate.proofs <- Conjunct 0 :: Assume :: Conjunct (n + 1) :: rest
    | _ -> E.s (E.bug "expecting conjunct in assume")
  in
  let finishConjuncts () =
    let rec nclosepar n = 
      if n = 0 then nil else text ")" ++ nclosepar (n - 1) 
    in
    match lfstate.proofs with
    | Conjunct n :: rest -> 
        lfstate.proofs <- rest;
        printLF 
          (fun () -> 
            (ignore (fprintf !proofChannel 
                       "truei%a" insert (nclosepar n))))
    | _ -> E.s (E.bug "expecting conjunct in finishConjuncts")
  in
  
  let unassume () = 
    (* Finish off the IMPI proof *)
    finishConjuncts ();
    ignore (fprintf !proofChannel ")@!");
    match lfstate.proofs with
      Assume :: rest -> lfstate.proofs <- rest
    | _ -> E.s (E.bug "Conjunct not finished")
  in

  let pick v = 
    ignore (fprintf !proofChannel
              "(alli [%s : _]@!" v)
  in
  let unpick () = 
    ignore (fprintf !proofChannel ")@!")
  in
     
  let finishVC () = 
    finishConjuncts ();
    ignore (fprintf !proofChannel "%%_LF)\n\n");
    match lfstate.proofs with
      [] -> ()
    | _ -> 
        ignore (E.warn "There are still %d un-retracted proofs in finishVC"
                  (List.length lfstate.proofs));
        List.iter (fun _ -> ignore (fprintf !proofChannel ")@!");) 
          lfstate.proofs;
        lfstate.proofs <- []
  in
  let conjunct (p: L.pred) (proof: P.proof) = 
    match lfstate.proofs with
      Conjunct n :: rest -> begin
        printLF 
          (fun () -> 
            ignore (fprintf !proofChannel "(andi @[%a@]@!" P.d_prf proof));
        lfstate.proofs <- Conjunct (n + 1) :: rest
      end
    | _ -> E.s (E.bug "expecting a conjunct")
  in
  let start () = () in
  let finish () = () in
  { Eng.startVC = startVC;
    Eng.start = start;
    Eng.finish = finish; 
    Eng.assume = assume;
    Eng.unassume = unassume;
    Eng.pick = pick;
    Eng.unpick = unpick;
    Eng.finishVC = finishVC;
    Eng.conjunct = conjunct;
    Eng.choose = 
        fun ~thechoice ~nrchoices -> E.s (E.bug "choose is not supported with LF")
  } 

(***** MAIN *****)  
let rec theMain () =
  let usageMsg = "Usage: kettle [options] lffiles" in
  let files : string list ref = ref [] in
  let recordFile fname = files := fname :: (!files) in
  let openLog lfile = 
    if !E.verboseFlag then
      ignore (Printf.printf "Setting log file to %s\n" lfile);
    try E.logChannel := open_out lfile with _ -> 
      raise (Arg.Bad "Cannot open log file") in
  let proofFile fname = 
    try proofChannel := open_out fname with _ ->
      raise (Arg.Bad "Cannot open proof file") in
  (* Initially all loaded procedures are enabled *)
  let includes : string list ref = ref [] in
  let enableProc name =
    if name = "ALL" then 
      includes := (List.map (fun p -> p.S.ident) allSatProcs)
    else begin
      if not (List.exists (fun sp -> sp.S.ident = name) allSatProcs) then
        raise (Arg.Bad ("Don't know anything about satproc " ^ name));
      includes := name :: !includes 
    end
  in
  enableProc "ALL";
  let disableProc name = 
    if name = "ALL" then 
      includes := []
    else begin
      if not (List.exists (fun sp -> sp.S.ident = name) allSatProcs) then
        raise (Arg.Bad ("Don't know anything about satproc " ^ name));
      includes := (List.filter (fun x -> x <> name) !includes)
    end
  in
  let importString : string ref = ref "" in
  let listProcs () = 
    ignore (E.log "Kettle uses satprocs:@!  @[EG   : main prover engine,@!%a@]\n"
        (docList ~sep:(chr ',' ++ line) (fun sp -> dprintf "%s   : %s"
            sp.S.ident sp.S.descr)) allSatProcs) in
  let setDebugFlag v name = 
    if name = "EG" then
      Eng.debug := v
    else
      let rec loop = function
          sp :: _ when sp.S.ident = name -> sp.S.debug := v
        | _ :: rest -> loop rest
        | [] -> E.s (E.error "Don't know about satproc %s\n" name)
      in
      loop allSatProcs
  in
  let axiomFiles : string list ref = ref [] in
  let setAxiomFile af = axiomFiles := af :: !axiomFiles in
  let useLFi = ref false in
  let usePerfCount = ref false in
  (*********** COMMAND LINE ARGUMENTS *****************)
  (* Prepare the UI options *)
  let uiOptions = 
    (* A way to collect argument options *)
    let rec mapConcat (f: 'a -> 'b list) (lst: 'a list) : 'b list = 
      match lst with 
        [] -> []
      | h :: t -> f h @ mapConcat f t
    in
    mapConcat
      (fun (ui: Eng.uiType) -> 
        let base = List.concat (List.map Options.optionToArgs ui#options) in
        if ui#name <> "" then begin
          ("-" ^ ui#name,
           Arg.Unit (fun _ -> Eng.theUI := ui),
           "use " ^ ui#name ^ " interface") ::
          base
        end else
          base)
      Features.uis
  in

  let argDescr = [
    "-listprocs", Arg.Unit listProcs, "list the loaded satprocs";
    "-debug", Arg.String (setDebugFlag true), 
                     "<xxx> turns on debugging for proc xxx";
    "-verbose", Arg.Unit (fun _ -> E.verboseFlag := true), 
                "turn on verbose mode";
    "-timing", Arg.Set Eng.timingFlag, 
                "print timing information, and reduce printing";
    "-nodebug", Arg.String (setDebugFlag false), 
                      "<xxx> turns off debugging for proc xxx";
    "-indent", Arg.Unit (fun _ -> indentProofs := true),
                "turn on indentation in printed proofs";
    "-log", Arg.String openLog, "the name of the log file";
    "-o", Arg.String proofFile, "the name of the output proof file";
    "-enable", Arg.String enableProc,  "name of satproc to enable (ALL for all loaded)";
    "-disable", Arg.String disableProc,  "name of satproc to disable (ALL for all loaded)";

    "-printinfix", Arg.Set Eng.printInfix,
                "print expressions using infix notation";
    "-printprefix", Arg.Clear Eng.printInfix,
                "print expressions using infix notation";
             
    "-lfi", Arg.Unit (fun x -> useLFi := true),
                "turn on LFi proof representation";
    "-perfcount", Arg.Unit (fun _ -> usePerfCount := true),
                 "use the performance counters";
    "-checkproofs", Arg.Unit (fun _ -> P.checkProofs := true),
                 "whether Kettle should check proofs";
    "-nostop", Arg.Unit (fun _ -> Eng.continue_on_error := true;
                                  Eng.continue_without_asking := true),
                 "continue after errors";
    "-pause", Arg.Unit (fun _ -> Eng.continue_on_error := true;
                                 Eng.continue_without_asking := false),
                 "continue after errors";
  ] @ Features.options 
    @ uiOptions
  in
  (* Add the parser features *)
  let selectedParser : S.parserSpec ref =
      (* Select the first one *)
      ref (match Features.allParsers with 
                 p :: _ -> p
               | _ -> E.s (E.bug "no parsers are present")) in
  let argDescr = 
    argDescr @
    (List.map 
       (fun p -> 
         ("-" ^ p.S.language ^ "input", 
          Arg.Unit (fun _ -> selectedParser := p), 
          "parse " ^ p.S.language ^ " input"))
       Features.allParsers)
  in 
  Arg.parse argDescr recordFile usageMsg;

  Stats.reset !usePerfCount;
  
  (* Grab the sat procs to be used. *)
  let satprocs = 
    List.filter (fun sp -> List.mem sp.S.ident !includes) 
      allSatProcs  
  in
  (* Now initialize the satprocs *)
  List.iter (fun sp -> sp.S.init ()) satprocs;

  ignore (E.log "Starting Kettle using satprocs: %a\n"
            (docList (fun sp -> text sp.S.ident)) satprocs);

  (* Construct the proof generator *)
  let proofgen = lfiProofGen in
  
  (* Construct the list of files to look at. Start with the axiom files *)
  files := (List.rev !axiomFiles) @ (List.rev !files);
  
  (* Print the proof preamble *)
  ignore (fprintf !proofChannel
            "ANN_STARTLF@!IMPORT %s @!ANN_ENDLF@!"
            !importString);
  
  let engstate = Eng.start satprocs proofgen in
  
  let rec processOneFile fname = 
    let lexbuf = E.startParsing fname in
    (* Now skip to the first LF token *)
    !selectedParser.S.initParser fname lexbuf;
    
    (* Write a function that serializes the result from the parser *)
    let inputs : I.input list ref = ref [] in
    let rec getNextInput () = 
      match !inputs with 
        i :: rest -> 
          inputs := rest; 
          (* ignore (E.log "Got %a\n" I.di i); *)
          i
      | [] -> (* Refresh *)
          inputs := !selectedParser.S.nextFewInputs lexbuf;
          assert (!inputs <> []);
          getNextInput ()
    in
    (* iterate over the toplevel forms in the file *)
    let rec doInputs () = 
      match getNextInput () with
        I.Axiom ax -> begin
          doInputs ()
        end
      | I.Import _ -> doInputs ()
      | I.Decl _ -> doInputs ()
      | I.Eof -> ()
      | I.TopChoose(i, nr) -> begin
          doInputs()
      end
      | I.StartGoal name -> begin  (* Now parse the elements of the goal *)
          (* Invoke the engine to prove this VC *)
          if !E.verboseFlag then 
            ignore (E.log "Proving goal %s\n" name);
          Stats.time "proveGoal" 
            (Eng.proveGoal name engstate) getNextInput;
          doInputs ();
      end
      | e -> E.s (E.bug "Input %a occurs at top-level" I.di e)
    in
    doInputs ();
    E.finishParsing ();
    Parsing.clear_parser ()
  in
  List.iter processOneFile !files;
  Eng.finish engstate;

  if !Eng.timingFlag || !E.verboseFlag then
    Stats.print stderr "Timings:\n";

  !Eng.theUI#run;

;;

(* Define here code to be called before termination *)
let cleanup () : unit = 
  if !E.logChannel != stderr then 
    close_out (! E.logChannel); 
  close_out (! proofChannel);
  ()
in

theMain ();
cleanup ();
let _= 
  if not !E.hadErrors then begin
    ignore (E.log "Kettle success\n");
    exit 0;
  end else begin
    ignore (E.log "Kettle failure\n");
    exit 1;
  end
in
()
;;
