(** The representation and checking of proofs *)
 
(* This module defines the language of proofs *)
val checkProofs: bool ref (** Whether to check the proofs *)

val printMagicProofWarning: bool ref
(** Whether to print warnings for Magic proofs *)


(* In order to maintain a compact representation of proofs we represent a 
 * proof node as the name of the axiom along with a list of arguments. An 
 * exception is made for the hypothetical and parametric proof nodes. *)
type parg = 
    I of int
  | S of string
  | E of Logic.exp
  | P of proof
  | F of Logic.pred

and proof = 
    Hyp of Util.symbol                    (* A hypothesis, 
                                           introduced with Impi, or Contra *) 
  | Truei
  | Falsee of proof 
    (** False elimination, if we have a proof of false *)

  | Andi of proof list
    (** Proves And[p1; p2; ... ] *)

  | Ande of int * proof
	(** Ande (i,pf) is a proof of p when
	   pf is a proof of And ps, where List.nth ps i = p *)

  | Alli of Logic.var * proof
  | Alle of  Logic.exp * proof

  | Existe of Logic.var * proof

  | Impi of Util.symbol * Logic.pred * proof (* bind var to a hypothetical 
                                            * proof of the pred. The last 
                                            * argument is a proof of the 
                                            * right hand side, possibly 
                                            * refering to var *)
  | Impe of proof * proof

                                        (* Bind var to a proof of the 
                                         * negation of pred. The third 
                                         * argument is a proof of false *)
  | Contra of Util.symbol * Logic.pred * proof
    
  | G of string * parg list             (* This is a generic proof 
                                         * constructor *)

  | Refl of Logic.exp 
  | Symm of proof
    (** Symm p proves e1=e2 if p:e2=e1 *)

  | Trans of Logic.exp * proof * proof
    (** Trans(et, p1t, p2t) proves e1=e2 if p1t:e1=et, pt2:et=e2 *)

  | Cong of Logic.exp * Logic.exp * proof list
  (** Cong \[pf\] proves Unop e1 = Unop e2 if pf proves e1 = e2;
      Cong \[pf1; pf2\] proves Binop (a1,b1) = Binop (a2,b2) if
        pf1 proves a1=a2 and pf2 proves b1=b2;
      Cong pfs proves Fun(f,as) = Fun(f,bs) if
        pfs.(n) proves as.(n) = bs.(n) 
     
      Note that each subproof of equality is really a proof 
      of (Var 0)=(Var 1) in the context in which (Var 0) and (Var 1)
      are replaced with the two things in their appropriate contexts.
   *)

  | WithLemma of Util.symbol (* The name of the lemma *)
                * Logic.exp list (* The expressions to instantiate the universally 
                            * quantified variables *)
                * proof list (* The proofs of the hypotheses *)

  | UnrollPDef of Logic.pred * proof (* Prove a predicate of the form 
                                * "PFun(n,args)" by replacing it with its 
                                * definition and providing a proof for the 
                                * resulting expansion *)

  | UnrollEDef of Logic.pred * string * proof
      (* Prove a predicate that uses an abbreviation expression constructor.
       * The given predicate should be the goal without the substitutions made.
       *)
  | RollEDef of Logic.pred * string * proof
      (* Like UnrollEDef, but proves the substituted version of the predicate *)

  | Arith of Logic.exp * Logic.exp         
   (** Check that e = e' by verifying that (e - e') mod32 = 0 using the rules 
    * of algebra. *)

  | ArithNeq of int * Logic.exp * Logic.exp   
    (** Check that e <> e' by verifiying that e - e' mod32 = i (i <> 0) using 
     * the rules of algebra *)

  | ArithApart of int * Logic.exp * Logic.exp 
     (** ArithApart(r, e1, e2) means that e1 - e2 = r and (-4 >= r or r >= 
      * 4).*)

  | ArithGeq of int * Logic.exp * Logic.exp  
        (** Check that e >= e', by verifying that e - e' = i and i >= 0. !!! 
         * This is unsound in presence of overflow. We'll have to clean this 
         * up. *)

  | ArithGt of int * Logic.exp * Logic.exp  
        (** Check that e >= e', by verifying that e - e' = i and i > 0. !!! 
         * This is unsound in presence of overflow. We'll have to clean this 
         * up. *)


  | SelUpd0 of Logic.exp * Logic.exp * Logic.exp 
        (* (SelUpd0 A V M) proves that (= V (sel A (upd A V M))) *)
  | SelUpd1 of Logic.exp * Logic.exp * proof
        (* (SelUpd1 V M D) proves that (= (sel A (upd A' V M)) (sel A M)) 
           provided that D proves A <> A' *)

  | Magic of string * Logic.pred 
    (** These are proofs that are allowed to go through the proof checker, 
     * with a warning. Pass a string to identify this magic *)

  | MagicLemma of lemma
    (** A magic proof of a lemma *)

  | Unimplemented of string     (* We'll expand the proof language as needed 
                                 * in the future. Use Unimplemented until 
                                 * then *)


and lemma = 
    { lname: Util.symbol; (* Lemma name *)
      lvars: string list; (* A list of variables that are universally 
                           * quantified over. These are bound in the rest of 
                           * the lemma, using UVar *)
      lhyps: Logic.pred list; (* A list of hypotheses. *)
      lconc: Logic.pred; (* A conclusion *)
    } 


   (* Infer the "type" of a proof. First argument is a list of variables and 
    * the second is a mapping of hypothesis variables to the predicates they 
    * prove. Raise ProofCheck exception if not a valid proof.  *)
val infer : Logic.var list -> 
            (Util.symbol * Logic.pred) list -> proof -> Logic.pred

val check : Logic.var list -> 
            (Util.symbol * Logic.pred) list -> proof -> Logic.pred -> unit



(**************** Lemmas ************)


(** Register a lemma, along with its proof *)
val registerLemma: 
    lemma -> 
    proof -> (* This is the proof that ALL lvars. (AND lhyps) => lconc *)
    unit

val registeredLemmas: unit -> lemma list

val findRegisteredLemma: Util.symbol -> lemma

(*************** Definitions **********)
(** Expression definitions. The name of the defined constructor (use a 
  * lowercase name), the list of formals, and the body, referring to the 
  * formals using deBruijn indices. *)
type edef = string * string list * Logic.exp 

(* Predicate definitions. The name of the defined constructor (use an 
 * uppercase initial for the name), the list of formals, and the body, 
 * referring to the formals using deBruijn indices. *)
type pdef = string * string list * Logic.pred 


(** Register an expression definition *)
val registerEDef: edef -> unit

(** Register a predicate definition *)
val registerPDef: pdef -> unit




 (** Make a proof by contradiction *)
val mkContraProof: Util.symbol (* H *) 
                -> Logic.pred (* A *)
                -> proof (* [H:pf (not A)] pf false *)
                -> proof (* pf A *)

val d_prf:  unit -> proof -> Pretty.doc
val dpa: parg -> Pretty.doc


(** We can install a new printer for the Proof module *)
val printer: Logic.logicPrinter ref

(** These are exported temporarily *)
val allEDefs: (string, edef) Hashtbl.t
val allPDefs: (string, pdef) Hashtbl.t
val allLemmas: lemma Inthash.t (* Indexed by symbols *)

(** Initialize the proof module *)
val init: unit -> unit


(** Take a snapshot of the state of the proof module. The only state is the 
 * set of registered lemmas and constructors *)
val snapshot: unit -> unit -> unit

val findLemma: lemma -> unit

val registeredLemmas: unit -> lemma list

val makeExistVar: string -> Logic.exp
