(* Tree Data Structure used to store Execution Path
 * of the VCGen
 * 
 * Node Structure: -> Stub Node (middle of execution)
 *             or  -> End Node  (finished execution)
 *
 * This tree representation assumes the following:
 * - Program is stored as a Sal Prog List (sal.ml)
 * - State is stored as a Sal Expression Array (sal.ml)
 * - Visited (labels) is stored as a String List
 * - Invariant (ivinfo) is stored as a invinfo List (stdvcgen.ml)
 * - Assumption is stored as a Sal Expression Stack (sal.ml)
 *
 *)
open Pretty
open Tk

module E = Errormsg
module H = Hashtbl

exception Empty

(* Redefine = and <> to support cyclic structures *)
let (=) = fun x y -> (Util.equals x y)
let (<>) = fun x y -> not (Util.equals x y)

(** Each element of the state tree must implement at least these methods *)
type state = <
  name: string;
  (** A name for the state, to be used in the UI *)

  infoline: string;
  (** A string to show on the info line *)

  filename: string option;
  (** A file to display *)

  line: int option;
  (** A line to set as current *)

  displayAssumptions: Pretty.doc list; 
  (** A list of assumptions **)

  displayState: Util.registerInfo list;
  (** A list of registers *)

  downcast: unit;
  (** Call this method to throw an exception that allows you to downcast this 
   * object *)

  copy: state;
  (** Make a copy of the state, for later use *)

>

(* The states in this tree is the nodeIndex, and the tree that grows *)
(* downwards from theRoot node *)

(* This implementation is a tree structure *)
and treeNode =  { 
    mutable parent   : treeNode option; (* None for a root node *)
    mutable state    : state;
    mutable children : treeNode list;
    mutable doneNode : bool;
    mutable index    : int;      (* index number to keep track of unique 
                                  * nodes  *)
    mutable collapsed: bool;  (* Whether this node is shown collapsed or not. 
                               * This makes sense only if it has children  *)
    mutable depth: int;       (** The depth in the tree. The root is at 
                                 * depth 0 *)
    mutable x: int;           (* The x coordinate of the canvas. This is 
                                 * the coordinate of the | *)
    mutable y: int;           (** The y coordinate of the canvas *)
  } 

type tree = { 
    mutable root           : treeNode;
    mutable nextIndex      : int;
            allNodes       : (int, treeNode) H.t;
    mutable currentNode    : treeNode option;
    mutable visibleNode:        treeNode option;
  } 


class rootState = object (self)
  method name = "Root"
  method infoline = "Root node"
  method filename: string option = None
  method line: int option = None
  method displayAssumptions: Pretty.doc list = []
  method displayState: Util.registerInfo list  = []
  method downcast: unit = 
    E.s (E.bug "You are calling downcast on the root node of the Stree")
  method copy = self
end


let nodeTemplate = 
  { parent = None;
    depth  = 0;
    state  = new rootState;
    children = [];
    doneNode = false;
    index    = 0;
    collapsed = false; x = 0; y = 0; }
    
(* Create a new node with a given parent *)
let newNode (t: tree) (s: state) (p: treeNode) : treeNode = 
  let n = { nodeTemplate with parent = Some p;
                              depth = 1 + p.depth;
                              state = s;
                              index = t.nextIndex } in
  (* Put it in the hash table *)
  H.add t.allNodes n.index n;
  t.nextIndex <- 1 + t.nextIndex;
  (* Now put it in the tree *)
  p.children <- p.children @ [ n ];
  n

(* create the new tree *)
let newTree (roots: state list) : tree =
  (* The root is a copy of the template *)
  let root = { nodeTemplate with index = 0 } in
  let rec t = 
    { root = root; 
      nextIndex = 1; (* Root has index 0 *) 
      allNodes = H.create 17; 
      currentNode = None;
      visibleNode = None 
    } 
  in

  List.iter (fun s -> ignore (newNode t s t.root)) roots;
  t
      

let currNode (t: tree) : treeNode = 
  match t.currentNode with
    None -> E.s (E.bug "stree:currNode is not set")
  | Some n -> n

let visibleNode (t: tree) : treeNode = 
  match t.visibleNode with
    None -> E.s (E.bug "stree:visibleNode is not set")
  | Some n -> n
  
        (* set the current state to an index. And make sure all of the 
         * parents are expanded.  *)
let setCurrent (t: tree) (node : treeNode) =
  t.currentNode <- Some node;
  t.visibleNode <- Some node;
  let rec loop = function
      None -> ()
    | Some p -> p.collapsed <- false; loop p.parent
  in
  loop (Some node)

        (* set the visible node. And make sure all of the parents are 
         * expanded. *)
let setVisible (t: tree) (node : treeNode) =
  t.visibleNode <- Some node;
  let rec loop = function
      None -> ()
    | Some p -> p.collapsed <- false; loop p.parent
  in
  loop (Some node)
      

(* getting next means going to next unprocessed node *)
let rec advanceCurrentToNextUnfinishedLeaf (t: tree) =
  let curr = currNode t in
  (* Get the parent *)
  let currp = 
    match curr.parent with 
      None -> E.s (E.bug "Stree: current node is the Root")
    | Some p -> p
  in
  
  (* Given a node, first the list of siblings that follow it among the 
   * children of its parent. *)
  let rec findRestSiblings (n: treeNode) = 
    let allSiblings = 
      match n.parent with 
        None -> E.s (E.bug "Stree: findRestSiblings for Root")
      | Some p -> p.children
    in
    let rec findStart = function
        [] -> E.s (E.bug "Cannot find node %d in its parent's children"
                     n.index)
      | (n' :: rest') as lst when n == n' -> rest'
      | _ :: rest' -> findStart rest'
    in
    findStart allSiblings
  in

  (** Mark a node done when all of its children are done *)
  let rec checkDone (n: treeNode) = 
    if not n.doneNode then begin
      if List.for_all (fun n' -> n'.doneNode) n.children then 
        n.doneNode <- true
    end;
    if n.doneNode && n.parent != None then 
      n.collapsed <- true
        
  in
  (** Find the first unfinished leaf in a list of siblings *)
  let rec findFirstLeaf (nl: treeNode list) (parent: treeNode) = 
    match nl with 
      [] -> begin
        if parent.index == t.root.index then
          raise Not_found
        else begin
          (* We have finished the scan for the children of this parent. But 
           * we might have started somewhere in the middle. Go and check 
           * whether we have done this one *)
          checkDone parent;
          (* Find the siblings of p *)
          let restSiblings = findRestSiblings parent in
          let pp = 
            match parent.parent with Some pp -> pp 
            | _ -> E.s (E.bug "Stree: not root and no parent")
          in
          findFirstLeaf restSiblings pp
        end
      end
    | n :: restn -> 
        if n.doneNode then 
          findFirstLeaf restn parent
        else
          (* If this is a leaf, then we are done *)
          if n.children == [] then begin
            setCurrent t n
          end else (* Not done, not a leaf. Go and look at the children. When 
                    * we finish those we'll come back and continue with the 
                    * restn  *)
            findFirstLeaf n.children n
  in
  let currRestSiblings = findRestSiblings curr in
  try
    (* We start searching at the current node *)
    findFirstLeaf (curr :: currRestSiblings) currp
  with Not_found -> begin
    (* We have not found it. Try one more time, from start *)
    try 
      findFirstLeaf t.root.children t.root
    with Not_found -> begin
      (* We are really done. Make the first child of the root to be the 
       * current node *)
      match t.root.children with 
        r :: _ -> setCurrent t r;
                  raise Not_found
      | _ -> E.s (E.bug "stree: tree has no nodes")
    end
  end

(* branch state is creating new node for each new state *)
let branchCurrentNode (t: tree) (newStates : state list) : unit =
  let n = 
    match t.currentNode with
      None -> E.s (E.bug "stree:advanceCurrent")
    | Some n -> n
  in
  assert (n.children = []);
  List.iter (fun s -> ignore (newNode t s n)) (List.rev newStates)


(* Interface to searching and returning a particular node *)
let root (t: tree) = t.root

(* returns the node with a particular index *)
let getNode (t: tree) (i : int) : treeNode =
  try
    H.find t.allNodes i
  with Not_found -> 
    E.s (E.bug "stree:getNode Not_found")
    

(* Iterate over all parents of a node, but ignore the real root, since that 
 * should not be visible by clients. *)
let rec iterParents (f: treeNode -> unit) (n: treeNode) : unit = 
  match n.parent with 
  | Some p when p.parent != None -> f p; iterParents f p
  | _ -> ()



(* Iterate over all the children of a node *)
let rec iterChildren (f: treeNode -> unit) (n: treeNode) : unit =
  List.iter (fun c -> f c; iterChildren f c) n.children

(* Iterate over the node AND all the children of a node *)
let rec iterNodeAndChildren (f: treeNode -> unit) (n: treeNode) : unit =
  f n; iterChildren f n

     
let deepCopy (t: tree) : tree = 
  
  (* make new tree fields *)
  let myAllNodes: (int, treeNode) H.t = H.create 17 in
  let myCurrentNode: treeNode option ref = ref None in
  
  (* lets do this the naive way.. copy recursively *)
  let rec copyList (po: treeNode option)
      (modelTree: treeNode list) : treeNode list =
    List.map
      (fun modeln ->
	copyNode po modeln)
      modelTree

  and copyNode (po: treeNode option) 
      (modelNode: treeNode) : treeNode = 
    let res = { modelNode with parent = po;
  		               state = modelNode.state#copy;
		               children = []; } in
    (* copy children *)
    res.children <- copyList (Some res) modelNode.children;
    (* add the node numbering *)
    H.add myAllNodes modelNode.index res;
    (* check if this is the current node *)
    if t.currentNode = (Some modelNode) then
      myCurrentNode := Some res;
    res
  in
  
  (* copy the tree *)
  let newRoot = copyNode None t.root in
  
  let newTree: tree = { root = newRoot;
		        nextIndex = t.nextIndex;
			allNodes = myAllNodes;
			currentNode = !myCurrentNode;
                        visibleNode = !myCurrentNode;
		      }
  in

  (* return the new tree *)
  newTree

