# A regression tester for Kettle
#
require 5.000;

# Packages to import.
use Getopt::Long;           # Command-line option processing
use File::Basename;         # File name parsing
use Cwd;                    # Directory navigation
use strict;
use Data::Dumper;
use FindBin;
use lib "$FindBin::Bin/../ocamlutil";

use RegTest;

print "Test infrastructure for Kettle\n";

# Create our customized test harness
my $TEST = KettleRegTest->new(AvailParams => { 'PROVE' => 1,
                                               'SUCCESS' => 0},
                              LogFile => "testkettle.log",
                              CommandName => "testkettle");

###
###
###
### Specialize RegTest
###
package KettleRegTest;

use strict;
use Data::Dumper;

BEGIN {
    use RegTest;
    @KettleRegTest::ISA = qw(RegTest);        # Inherit from RegTest
}

# The constructor
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@_);

    # print Dumper($self);

    return $self;
}



sub extraOptions {
    my($self) = @_;
    return "";
}

sub extraHelpMessage {
    my($self) = @_;
    
    my ($scriptname, $extra) = $self->SUPER::extraHelpMessage();
    return ("testkettle",
            $extra . << "EOF");

  Default log file is testkettle.log
EOF
}

sub errorHeading {
    my($self, $err) = @_;
    return "Not executed" if $err == -1;
    return "Success" if $err == 0;
    return "Initial" if $err == 1000;
    return "Parse error" if $err == 1001;
    return "Proving error" if $err == 1002;
    return $self->SUPER::errorHeading($err);
}

sub startParsingLog {
    my($self, $tst) = @_;
    $tst->{instage} = 1000;
    $tst->{ErrorCode} = 0;
}


sub availableParameters {
    my($self) = @_;
    return %::availpars;
}

# We watch the log and we remember in what stage we are (so that we can
# interpret the error)

# Stages:
#  1000 - Start
#  1001 - Parsing
#  1002 - Proving

my %commonerrors = 
    ("^Parsing file" => sub { $_[1]->{instage} = 1001; },

     "^Proving goal with ([\\d]+) points" 
     => sub { $_[1]->{instage} = 1002; 
              $_[1]->{POINTS}  = $_[2]; },
     "^Proving goal" 
     => sub { $_[1]->{instage} = 1002; },
# We are seeing an error from make. Try to classify it based on the stage
# in which we are
     "^make: \\*\\*\\*" => 
     sub { 
         if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},

    "^Segmentation fault" =>
     sub {
        if($_[1]->{ErrorCode} == 0) {
             $_[1]->{ErrorCode} = $_[1]->{instage};
         }},    

     "^Ran ([\\d.]+) iterations" => sub { $_[1]->{ITERS} = $_[2]; },
     
     "^Time\\s+([\\d.]+) ms" => sub { $_[1]->{PROVE} = $_[2]; },
     "^Error parsing:" => sub { $_[1]->{ErrorCode} = 1001; },
     "^Error .+:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; },
     "^Bug:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; 
                      if(! defined $_[1]->{ErrorMsg}) {
                          $_[1]->{ErrorMsg} = $_; 
                      }},
     "^Unimplemented:" => sub { $_[1]->{ErrorCode} = $_[1]->{instage}; 
                      if(! defined $_[1]->{ErrorMsg}) {
                          $_[1]->{ErrorMsg} = $_; 
                      }},
    );

                                         

sub addTest {
    my($name, # The name of the file (withouth the extension)
       %args, # An optional list of key => data
              #   Group => ['grp1', 'grp2' ]
              #   Enabled => 0
              #   ExtraArgs => "something to add to command line"
              #   ExtraKettleArgs => "something to add to Kettle command line"
              #   Comm => "a comment to be printed if it fails"
              #   ExpectPattern => "something to appear on output"
              #   MustFail => 1
       ) = @_;

    my $tst = 
        $TEST->newTest(
                  Name => $name,
                  Dir => ".",
                  Cmd => "make EXTRAARGS=\"-checkproofs $args{ExtraKettleArgs}\" $name",
                  Group => [],
                  ExtraArgs => $args{ExtraArgs},
                  Patterns => \%commonerrors);
    delete $args{ExtraKettleArgs};
    foreach my $k (keys %args) {
        $tst->{$k} = $args{$k};
    }
    return $tst;
}

# test/equality 
foreach my $t ("0", "1", "2", "3", "4", "5", "7", "1_SATb", 
               "uninterpreted") {
    &addTest("prove/equality/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC");
}
       
foreach my $t ("uf0-true", "uf1-true", "uf-single", "uf-copies", 
               "uf-wrong", "uf2-false", "uf2-true", 
               "uf3-true", "uf3-false") {
    &addTest("prove/sumit/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC");
}         

foreach my $t ("uf1-false", "uf2-false", "uf3-false") {
    &addTest("prove/sumit/" . $t,  
             Group => ['cc'],
             ExtraKettleArgs => "-enable CC",
             MustFail => 1);
}         

foreach my $t ("0", "1", "2", "3", "4", "5", "6", "7", 
               "8", "9", "10", "11", "12", "13", "14", "15", 
               "16", "17", "18", "19", "20", "21", 
               "40", "greg", "hypax", "neq0", "unint",
               "w1vc") {
    &addTest("prove/cmu/" . $t,  
             Group => ['cc', 'simplex'],
             ExtraKettleArgs => "-enable CC -enable SX");
}         


foreach my $t ("simplex/Simplex") 
{
    &addTest("prove/tcj/" . $t, 
             Group => ['tcj'],
             ExtraArgs => "USE_TCJ=1");
}


$TEST->addComment("prove/cmu/21", "parsing => as part of the goal");
$TEST->addComment("prove/cmu/hypax", 
                  "parsing complex pred as part of assumption");
$TEST->addComment("prove/cmu/12", 
                  "does not appear to be provable!");
$TEST->getTest("prove/cmu/7")->{MustFail} = 1;
$TEST->getTest("prove/cmu/9")->{MustFail} = 1;

foreach my $t ("1", "2", "3", "4", "5", "6", "7", "8", "9",
               "simple0", "simple1", "simple2", "simple3", "simple4" ) {
    &addTest("prove/arith/" . $t,  
             Group => ['cc', 'simplex'],
             ExtraArgs => "-enable CC -enable SX");
}
$TEST->addComment("prove/arith/5", "\\/ in left-side of implication");


&addTest("prove/arith/pfbug1", 
         ExtraKettleArgs => "-nostop",
         MustFail => 1,
         ExpectPattern => "foo",
         Group => ['simplex']);

&addTest("prove/array1");
&addTest("prove/arrays",
         Comm => "arrays are not complete");

#####################################################



# Now invoke it
$TEST->doit();

# print Dumper($TEST);

exit(0);


1;



